#include "Build.h"

void getBuild(Value* pValue, BuildResult* pResult,
              int slotWithZero, int slotWithoutZero)
{
    const int fast = 1;
    const int slow = 0;
    const int withZero = 1;
    const int withoutZero = 0;

    switch (pValue->m_type)
    {
    case VT_int:
    {
        int i = pValue->m_int;
        int turns = 1;
        while (i > 0)
        {
            Card card;
            if (i & 1)
            {
                i--;
                card = succ;
            }
            else
            {
                i >>= 1;
                card = dbl;
            }
            turns++;

            pResult->m_moves[fast][withoutZero].insert(
                pResult->m_moves[fast][withoutZero].begin(),
                Move(card, slotWithoutZero));

            pResult->m_moves[fast][withZero].insert(
                pResult->m_moves[fast][withZero].begin(),
                Move(card, slotWithZero));

            pResult->m_moves[slow][withoutZero].push_back(
                Move(K, slotWithoutZero));
            pResult->m_moves[slow][withoutZero].push_back(
                Move(S, slotWithoutZero));
            pResult->m_moves[slow][withoutZero].push_back(
                Move(slotWithoutZero, card));

            pResult->m_moves[slow][withZero].insert(
                pResult->m_moves[slow][withZero].begin(),
                Move(card, slotWithoutZero));
        }

        pResult->m_moves[fast][withoutZero].insert(
            pResult->m_moves[fast][withoutZero].begin(),
            Move(slotWithoutZero, zero));

        pResult->m_moves[fast][withZero].insert(
            pResult->m_moves[fast][withZero].begin(),
            Move(slotWithZero, zero));

        pResult->m_moves[slow][withoutZero].push_back(
            Move(slotWithoutZero, zero));

        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(slotWithoutZero, zero));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(put, slotWithoutZero));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(slotWithZero, get));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(S, slotWithZero));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(K, slotWithZero));
        pResult->m_moves[slow][withZero].push_back(
            Move(slotWithZero, zero));

        pResult->m_turns[fast][withoutZero] = turns;
        pResult->m_turns[fast][withZero] = turns;
        pResult->m_usesZero[fast] = false;

        pResult->m_turns[slow][withoutZero] = turns * 3 - 2;
        pResult->m_turns[slow][withZero] = turns + 5;

        if (pResult->m_turns[slow][withZero] <
            pResult->m_turns[slow][withoutZero])
        {
            pResult->m_usesZero[slow] = true;
        }
        else
        {
            pResult->m_usesZero[slow] = false;
            pResult->m_turns[slow][withZero] =
                pResult->m_turns[slow][withoutZero];
        }
    }
    break;
    case VT_I:
    case VT_succ:
    case VT_dbl:
    case VT_get:
    case VT_put:
    case VT_S:
    case VT_K:
    case VT_inc:
    case VT_dec:
    case VT_attack:
    case VT_help:
    case VT_copy:
    case VT_revive:
    case VT_zombie:
    {
        Card card;
        switch (pValue->m_type)
        {
        case VT_I: card = I; break;
        case VT_succ: card = succ; break;
        case VT_dbl: card = dbl; break;
        case VT_get: card = get; break;
        case VT_put: card = put; break;
        case VT_S: card = S; break;
        case VT_K: card = K; break;
        case VT_inc: card = inc; break;
        case VT_dec: card = dec; break;
        case VT_attack: card = attack; break;
        case VT_help: card = help; break;
        case VT_copy: card = copy; break;
        case VT_revive: card = revive; break;
        case VT_zombie: card = zombie; break;
        }

        pResult->m_turns[fast][withoutZero] = 1;
        pResult->m_moves[fast][withoutZero].resize(1);
        pResult->m_moves[fast][withoutZero][0] =
            Move(slotWithoutZero, card);

        pResult->m_turns[fast][withZero] = 1;
        pResult->m_usesZero[fast] = false;
        pResult->m_moves[fast][withZero].resize(1);
        pResult->m_moves[fast][withZero][0] =
            Move(slotWithZero, card);

        pResult->m_turns[slow][withoutZero] = 1;
        pResult->m_moves[slow][withoutZero].resize(1);
        pResult->m_moves[slow][withoutZero][0] =
            Move(slotWithoutZero, card);

        pResult->m_turns[slow][withZero] = 1;
        pResult->m_usesZero[slow] = false;
        pResult->m_moves[slow][withZero].resize(1);
        pResult->m_moves[slow][withZero][0] =
            Move(slotWithZero, card);
    }
    break;
    case VT_S_1:
    case VT_K_1:
    case VT_attack_1:
    case VT_help_1:
    case VT_zombie_1:
    {
        Card card;
        switch (pValue->m_type)
        {
        case VT_S_1: card = S; break;
        case VT_K_1: card = K; break;
        case VT_attack_1: card = attack; break;
        case VT_help_1: card = help; break;
        case VT_zombie_1: card = zombie; break;
        }

        BuildResult result1;
        getBuild(pValue->m_arg1, &result1, slotWithZero, slotWithoutZero);

        pResult->m_turns[fast][withoutZero] =
            1 + result1.m_turns[fast][withoutZero];
        pResult->m_turns[fast][withZero] =
            1 + result1.m_turns[fast][withZero];
        pResult->m_usesZero[fast] = result1.m_usesZero[fast];

        pResult->m_turns[slow][withoutZero] =
            3 + result1.m_turns[slow][withoutZero];
        pResult->m_turns[slow][withZero] =
            3 + result1.m_turns[slow][withZero];
        pResult->m_usesZero[slow] = result1.m_usesZero[slow];

        pResult->m_moves[fast][withoutZero] =
            result1.m_moves[fast][withoutZero];
        pResult->m_moves[fast][withoutZero].push_back(
            Move(card, slotWithoutZero));

        pResult->m_moves[fast][withZero] =
            result1.m_moves[fast][withZero];
        pResult->m_moves[fast][withZero].push_back(
            Move(card, slotWithZero));

        pResult->m_moves[slow][withoutZero] =
            result1.m_moves[slow][withoutZero];
        pResult->m_moves[slow][withoutZero].insert(
            pResult->m_moves[slow][withoutZero].begin(),
            Move(slotWithoutZero, card));
        pResult->m_moves[slow][withoutZero].insert(
            pResult->m_moves[slow][withoutZero].begin(),
            Move(S, slotWithoutZero));
        pResult->m_moves[slow][withoutZero].insert(
            pResult->m_moves[slow][withoutZero].begin(),
            Move(K, slotWithoutZero));

        pResult->m_moves[slow][withZero] =
            result1.m_moves[slow][withZero];
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(slotWithZero, card));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(S, slotWithZero));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].begin(),
            Move(K, slotWithZero));
    }
    break;
    case VT_S_2:
    case VT_attack_2:
    case VT_help_2:
    {
        Card card;
        switch (pValue->m_type)
        {
        case VT_S_2: card = S; break;
        case VT_attack_2: card = attack; break;
        case VT_help_2: card = help; break;
        }

        BuildResult result1;
        getBuild(pValue->m_arg1, &result1, slotWithZero, slotWithoutZero);

        BuildResult result2;
        getBuild(pValue->m_arg2, &result2, slotWithZero, slotWithoutZero);

        pResult->m_turns[fast][withoutZero] =
            1 +
            result1.m_turns[fast][withoutZero] +
            result2.m_turns[slow][withoutZero];

        pResult->m_moves[fast][withoutZero] =
            result1.m_moves[fast][withoutZero];
        pResult->m_moves[fast][withoutZero].push_back(
            Move(card, slotWithoutZero));
        pResult->m_moves[fast][withoutZero].insert(
            pResult->m_moves[fast][withoutZero].end(),
            result2.m_moves[slow][withoutZero].begin(),
            result2.m_moves[slow][withoutZero].end());

        pResult->m_moves[fast][withZero] =
            result1.m_moves[fast][withZero];
        pResult->m_moves[fast][withZero].push_back(
            Move(card, slotWithZero));
        if (result2.m_turns[fast][withoutZero] + 5 <
            result2.m_turns[slow][withZero])
        {
            pResult->m_turns[fast][withZero] =
                1 +
                result1.m_turns[fast][withZero] +
                result2.m_turns[fast][withoutZero] + 5;
            pResult->m_usesZero[fast] = true;

            pResult->m_moves[fast][withZero].push_back(
                Move(K, slotWithZero));
            pResult->m_moves[fast][withZero].push_back(
                Move(S, slotWithZero));
            pResult->m_moves[fast][withZero].push_back(
                Move(slotWithZero, get));
            pResult->m_moves[fast][withZero].push_back(
                Move(put, slotWithoutZero));
            pResult->m_moves[fast][withZero].insert(
                pResult->m_moves[fast][withZero].end(),
                result2.m_moves[fast][withoutZero].begin(),
                result2.m_moves[fast][withoutZero].end());
            pResult->m_moves[fast][withZero].push_back(
                Move(slotWithZero, zero));
        }
        else
        {
            pResult->m_turns[fast][withZero] =
                1 +
                result1.m_turns[fast][withZero] +
                result2.m_turns[slow][withZero];
            pResult->m_usesZero[fast] =
                result1.m_usesZero[fast] ||
                result2.m_usesZero[slow];

            pResult->m_moves[fast][withZero].insert(
                pResult->m_moves[fast][withZero].end(),
                result2.m_moves[slow][withZero].begin(),
                result2.m_moves[slow][withZero].end());
        }

        pResult->m_turns[slow][withoutZero] =
            5 +
            result1.m_turns[slow][withoutZero] +
            result2.m_turns[slow][withoutZero];

        pResult->m_moves[slow][withoutZero].push_back(
            Move(K, slotWithoutZero));
        pResult->m_moves[slow][withoutZero].push_back(
            Move(S, slotWithoutZero));
        pResult->m_moves[slow][withoutZero].push_back(
            Move(K, slotWithoutZero));
        pResult->m_moves[slow][withoutZero].push_back(
            Move(S, slotWithoutZero));
        pResult->m_moves[slow][withoutZero].push_back(
            Move(slotWithoutZero, card));
        pResult->m_moves[slow][withoutZero].insert(
            pResult->m_moves[slow][withoutZero].end(),
            result1.m_moves[slow][withoutZero].begin(),
            result1.m_moves[slow][withoutZero].end());
        pResult->m_moves[slow][withoutZero].insert(
            pResult->m_moves[slow][withoutZero].end(),
            result2.m_moves[slow][withoutZero].begin(),
            result2.m_moves[slow][withoutZero].end());

        pResult->m_moves[slow][withZero].push_back(
            Move(K, slotWithZero));
        pResult->m_moves[slow][withZero].push_back(
            Move(S, slotWithZero));
        pResult->m_moves[slow][withZero].push_back(
            Move(K, slotWithZero));
        pResult->m_moves[slow][withZero].push_back(
            Move(S, slotWithZero));
        pResult->m_moves[slow][withZero].push_back(
            Move(slotWithZero, card));
        pResult->m_moves[slow][withZero].insert(
            pResult->m_moves[slow][withZero].end(),
            result1.m_moves[slow][withZero].begin(),
            result1.m_moves[slow][withZero].end());
        if (result2.m_turns[fast][withoutZero] + 5 <
            result2.m_turns[slow][withZero])
        {
            pResult->m_turns[slow][withZero] =
                5 +
                result1.m_turns[slow][withZero] +
                result2.m_turns[fast][withoutZero] + 5;
            pResult->m_usesZero[slow] = true;

            pResult->m_moves[slow][withZero].push_back(
                Move(K, slotWithZero));
            pResult->m_moves[slow][withZero].push_back(
                Move(S, slotWithZero));
            pResult->m_moves[slow][withZero].push_back(
                Move(slotWithZero, get));
            pResult->m_moves[slow][withZero].push_back(
                Move(put, slotWithoutZero));
            pResult->m_moves[slow][withZero].insert(
                pResult->m_moves[slow][withZero].end(),
                result2.m_moves[fast][withoutZero].begin(),
                result2.m_moves[fast][withoutZero].end());
            pResult->m_moves[slow][withZero].push_back(
                Move(slotWithZero, zero));
        }
        else
        {
            pResult->m_turns[slow][withZero] =
                5 +
                result1.m_turns[slow][withZero] +
                result2.m_turns[slow][withZero];
            pResult->m_usesZero[slow] =
                result1.m_usesZero[slow] ||
                result2.m_usesZero[slow];

            pResult->m_moves[slow][withZero].insert(
                pResult->m_moves[slow][withZero].end(),
                result2.m_moves[slow][withZero].begin(),
                result2.m_moves[slow][withZero].end());
        }
    }
    break;
    }
}

Value* _(Card card)
{
    return new Value(card);
}

Value* _(int i)
{
    Value* pResult = new Value();
    pResult->m_type = VT_int;
    pResult->m_int = i;
    return pResult;
}

Value* _(Value* pValue)
{
    return pValue;
}
