#include "Program.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <vector>

using std::vector;

bool rd(Move* pMove)
{
    Move move;

    const int BUFSIZE = 1024;
    char buf[BUFSIZE];

    if (!fgets(buf, BUFSIZE, stdin))
    {
        //fprintf(stderr, "Read failed\n");
        //exit(1);
        //exit(0);
        return false;
    }

    if (strcmp(buf, "1\n") == 0)
    {
        move.m_apply = Left;
    }
    else if (strcmp(buf, "2\n") == 0)
    {
        move.m_apply = Right;
    }
    else
    {
        fprintf(stderr, "Bad input\n");
        exit(1);
    }

    for (int i = 0; i < 2; i++)
    {
        if (!fgets(buf, BUFSIZE, stdin))
        {
            //fprintf(stderr, "Read failed\n");
            //exit(1);
            //exit(0);
            return false;
        }

        int j = (move.m_apply == Left) ? i : 1-i;

        if (j == 0)
        {
            if (strcmp(buf, "I\n") == 0) { move.m_card = I; }
            else if (strcmp(buf, "zero\n") == 0) { move.m_card = zero; }
            else if (strcmp(buf, "succ\n") == 0) { move.m_card = succ; }
            else if (strcmp(buf, "dbl\n") == 0) { move.m_card = dbl; }
            else if (strcmp(buf, "get\n") == 0) { move.m_card = get; }
            else if (strcmp(buf, "put\n") == 0) { move.m_card = put; }
            else if (strcmp(buf, "S\n") == 0) { move.m_card = S; }
            else if (strcmp(buf, "K\n") == 0) { move.m_card = K; }
            else if (strcmp(buf, "inc\n") == 0) { move.m_card = inc; }
            else if (strcmp(buf, "dec\n") == 0) { move.m_card = dec; }
            else if (strcmp(buf, "attack\n") == 0) { move.m_card = attack; }
            else if (strcmp(buf, "help\n") == 0) { move.m_card = help; }
            else if (strcmp(buf, "copy\n") == 0) { move.m_card = copy; }
            else if (strcmp(buf, "revive\n") == 0) { move.m_card = revive; }
            else if (strcmp(buf, "zombie\n") == 0) { move.m_card = zombie; }
            else
            {
                fprintf(stderr, "Bad input\n");
                exit(1);
            }
        }
        else
        {
            move.m_slot = atoi(buf);
            if (move.m_slot < 0 || move.m_slot > 255)
            {
                fprintf(stderr, "Bad input\n");
                exit(1);
            }
        }
    }

    if (pMove)
        *pMove = move;

    return true;
}

void wr(Move move)
{
    if (move.m_apply == Left)
    {
        printf("1\n");
    }
    else if (move.m_apply == Right)
    {
        printf("2\n");
    }
    else
    {
        fprintf(stderr, "Bad output\n");
        exit(1);
    }

    for (int i = 0; i < 2; i++)
    {
        int j = (move.m_apply == Left) ? i : 1-i;

        if (j == 0)
        {
            switch (move.m_card)
            {
            case I: printf("I\n"); break;
            case zero: printf("zero\n"); break;
            case succ: printf("succ\n"); break;
            case dbl: printf("dbl\n"); break;
            case get: printf("get\n"); break;
            case put: printf("put\n"); break;
            case S: printf("S\n"); break;
            case K: printf("K\n"); break;
            case inc: printf("inc\n"); break;
            case dec: printf("dec\n"); break;
            case attack: printf("attack\n"); break;
            case help: printf("help\n"); break;
            case copy: printf("copy\n"); break;
            case revive: printf("revive\n"); break;
            case zombie: printf("zombie\n"); break;
            default:
                fprintf(stderr, "Bad output\n");
                exit(1);
            }
        }
        else
        {
            if (move.m_slot < 0 || move.m_slot > 255)
            {
                fprintf(stderr, "Bad input\n");
                exit(1);
            }

            printf("%d\n", move.m_slot);
        }
    }

    fflush(stdout);

    //fprintf(stderr, "################## HELLO! ###################\n");
    //fflush(stderr);
}

Program::Program() :
    m_player(0),
    m_turn(1)
{
}

Program::~Program()
{
}

int Program::run(int argc, char* argv[])
{
    if (argc != 2)
    {
        fprintf(stderr, "Bad arguments\n");
        return 1;
    }

    if (strcmp(argv[1], "0") == 0)
    {
        m_player = 0;
    }
    else if (strcmp(argv[1], "1") == 0)
    {
        m_player = 1;
    }
    else
    {
        fprintf(stderr, "Bad arguments\n");
        return 1;
    }

    srand(time(0) ^ (m_player == 0 ? 0xdeadbeef : 0xfeedface));

    FILE* logFile = 0;
#if 1
    {
        char logFileName[65536];
        sprintf(logFileName, "log%d.txt", m_player);
        logFile = fopen(logFileName, "w");
    }
#endif

    Move move;

    bool first = true;
    while (true)
    {
        if (m_turn > 100000)
        {
            break;
        }

        if (logFile)
        {
            fprintf(logFile, "\n");
            fprintf(logFile, "###### turn %d\n", m_turn);
            fflush(logFile);
        }

        bool alive0 = false;
        bool alive1 = false;

        for (int i = 0; i < 2; i++)
        {
            int j = (m_player == 0) ? i : 1-i;

            if (logFile)
            {
                fprintf(logFile, "\n");
                m_state.dump(logFile);
                fflush(logFile);
            }

            vector<int> zombies;
            m_state.preStep(logFile, &zombies);

            if (j == 0)
            {
                opponentZombies(zombies);
                proponent(&move);
                wr(move);
                m_state.step(logFile, move);
            }
            else
            {
                proponentZombies(zombies);
                if (!rd(&move))
                {
                    break;
                }
                m_state.step(logFile, move);
                opponent(move);
            }

            alive0 = false;
            alive1 = false;
            for (int slot = 0; slot < 256; slot++)
            {
                if (m_state.m_v[0][slot] > 0)
                {
                    alive0 = true;
                }
                if (m_state.m_v[1][slot] > 0)
                {
                    alive1 = true;
                }
                if (alive0 && alive1)
                {
                    break;
                }
            }
            if (!alive0 || !alive1)
            {
                break;
            }
        }

        if (!alive0 || !alive1)
        {
            break;
        }

        m_turn++;
    }

    if (logFile)
    {
        fclose(logFile);
    }

    return 0;
}

void Program::opponent(Move move)
{
}

void Program::opponentZombies(std::vector<int>& zombies)
{
}

void Program::proponentZombies(std::vector<int>& zombies)
{
}
