#include "State.h"

#include <vector>

using std::vector;

int Value::s_allocCount = 0;

Value::Value()
    : m_type(VT_I), m_arg1(0), m_arg2(0), m_int(0), m_refCount(1)
{
    s_allocCount++;
    //fprintf(stderr, "s_allocCount = %d\n", s_allocCount);
}

Value::Value(Card card)
    : m_type(VT_I), m_arg1(0), m_arg2(0), m_int(0), m_refCount(1)
{
    s_allocCount++;
    //fprintf(stderr, "s_allocCount = %d\n", s_allocCount);

    switch (card)
    {
    case I: m_type = VT_I; break;
    case zero: m_type = VT_int; m_int = 0; break;
    case succ: m_type = VT_succ; break;
    case dbl: m_type = VT_dbl; break;
    case get: m_type = VT_get; break;
    case put: m_type = VT_put; break;
    case S: m_type = VT_S; break;
    case K: m_type = VT_K; break;
    case inc: m_type = VT_inc; break;
    case dec: m_type = VT_dec; break;
    case attack: m_type = VT_attack; break;
    case help: m_type = VT_help; break;
    case copy: m_type = VT_copy; break;
    case revive: m_type = VT_revive; break;
    case zombie: m_type = VT_zombie; break;
    }
}

Value::~Value()
{
    if (m_arg1)
    {
        m_arg1->release();
    }

    if (m_arg2)
    {
        m_arg2->release();
    }

    s_allocCount--;
    //fprintf(stderr, "s_allocCount = %d\n", s_allocCount);
}

bool Value::equals(const Value* pOther) const
{
    if (this == pOther)
    {
        return true;
    }

    if (m_type != pOther->m_type)
    {
        return false;
    }

    switch (m_type)
    {
    case VT_int:
    {
        if (m_int != pOther->m_int)
        {
            return false;
        }
    }
    break;
    case VT_I:
    case VT_succ:
    case VT_dbl:
    case VT_get:
    case VT_put:
    case VT_S:
    case VT_K:
    case VT_inc:
    case VT_dec:
    case VT_attack:
    case VT_help:
    case VT_copy:
    case VT_revive:
    case VT_zombie:
    {
    }
    break;
    case VT_S_1:
    case VT_K_1:
    case VT_attack_1:
    case VT_help_1:
    case VT_zombie_1:
    {
        if (!m_arg1->equals(pOther->m_arg1))
        {
            return false;
        }
    }
    break;
    case VT_S_2:
    case VT_attack_2:
    case VT_help_2:
    {
        if (!m_arg1->equals(pOther->m_arg1) ||
            !m_arg2->equals(pOther->m_arg2))
        {
            return false;
        }
    }
    break;
    }

    return true;
}

void Value::dump(FILE* file) const
{
    switch (m_type)
    {
    case VT_int:
    {
        if (m_int == 0)
        {
            fprintf(file, "zero");
        }
        else
        {
            fprintf(file, "%d", m_int);
        }
    }
    break;
    case VT_I:
    {
        fprintf(file, "I");
    }
    break;
    case VT_succ:
    {
        fprintf(file, "succ");
    }
    break;
    case VT_dbl:
    {
        fprintf(file, "dbl");
    }
    break;
    case VT_get:
    {
        fprintf(file, "get");
    }
    break;
    case VT_put:
    {
        fprintf(file, "put");
    }
    break;
    case VT_S:
    {
        fprintf(file, "S");
    }
    break;
    case VT_S_1:
    {
        fprintf(file, "S(");
        m_arg1->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_S_2:
    {
        fprintf(file, "S(");
        m_arg1->dump(file);
        fprintf(file, ")(");
        m_arg2->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_K:
    {
        fprintf(file, "K");
    }
    break;
    case VT_K_1:
    {
        fprintf(file, "K(");
        m_arg1->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_inc:
    {
        fprintf(file, "inc");
    }
    break;
    case VT_dec:
    {
        fprintf(file, "dec");
    }
    break;
    case VT_attack:
    {
        fprintf(file, "attack");
    }
    break;
    case VT_attack_1:
    {
        fprintf(file, "attack(");
        m_arg1->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_attack_2:
    {
        fprintf(file, "attack(");
        m_arg1->dump(file);
        fprintf(file, ")(");
        m_arg2->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_help:
    {
        fprintf(file, "help");
    }
    break;
    case VT_help_1:
    {
        fprintf(file, "help(");
        m_arg1->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_help_2:
    {
        fprintf(file, "help(");
        m_arg1->dump(file);
        fprintf(file, ")(");
        m_arg2->dump(file);
        fprintf(file, ")");
    }
    break;
    case VT_copy:
    {
        fprintf(file, "copy");
    }
    break;
    case VT_revive:
    {
        fprintf(file, "revive");
    }
    break;
    case VT_zombie:
    {
        fprintf(file, "zombie");
    }
    break;
    case VT_zombie_1:
    {
        fprintf(file, "zombie(");
        m_arg1->dump(file);
        fprintf(file, ")");
    }
    break;
    }
}

State::State()
    : m_player(0),
      m_turn(1),
      m_zombie(false)
{
    m_pI = new Value();
    m_pI->m_type = VT_I;

    for (int player = 0; player < 2; player++)
    {
        for (int slot = 0; slot < 256; slot++)
        {
            m_v[player][slot] = 10000;
            m_pI->addRef();
            m_f[player][slot] = m_pI;
        }
    }
}

State::~State()
{
    for (int player = 0; player < 2; player++)
    {
        for (int slot = 0; slot < 256; slot++)
        {
            m_f[player][slot]->release();
        }
    }
    m_pI->release();
}

State::State(const State& other)
    : m_player(other.m_player),
      m_turn(other.m_turn)
{
    for (int player = 0; player < 2; player++)
    {
        for (int slot = 0; slot < 256; slot++)
        {
            m_v[player][slot] = other.m_v[player][slot];
            m_f[player][slot] = other.m_f[player][slot];
            m_f[player][slot]->addRef();
        }
    }
}

State& State::operator=(const State& other)
{
    if (this != &other)
    {
        m_player = other.m_player;
        m_turn = other.m_turn;
        for (int player = 0; player < 2; player++)
        {
            for (int slot = 0; slot < 256; slot++)
            {
                m_v[player][slot] = other.m_v[player][slot];
                m_f[player][slot]->release();
                m_f[player][slot] = other.m_f[player][slot];
                m_f[player][slot]->addRef();
            }
        }
    }

    return *this;
}

class Error {};

void State::preStep(FILE* file, vector<int>* pZombies)
{
    m_zombie = true;

    pZombies->clear();

    for (int slot = 0; slot < 256; slot++)
    {
        if (m_v[m_player][slot] == -1)
        {
            pZombies->push_back(slot);

            if (file)
            {
                fprintf(file, "Zombie move: slot %d -> I\n", slot);
                fflush(file);
            }

            Value* pFunc = m_f[m_player][slot];
            pFunc->addRef();
            Value* pArg = m_pI;
            pArg->addRef();

            try
            {
                m_applyCount = 0;
                Value* pResult = apply(file, pFunc, pArg);
                pResult->release();
            }
            catch (Error)
            {
            }

            m_f[m_player][slot]->release();
            m_f[m_player][slot] = m_pI;
            m_pI->addRef();

            pFunc->release();
            pArg->release();

            m_v[m_player][slot] = 0;
        }
    }

    m_zombie = false;
}

void State::step(FILE* file, Move move)
{
    if (file)
    {
        fprintf(file, "Move: ");
        move.dump(file);
        fprintf(file, "\n");
        fflush(file);
    }

    Value* pCardValue = new Value(move.m_card);
    Value* pSlotValue = m_f[m_player][move.m_slot];
    pSlotValue->addRef();

    Value* pFunc = (move.m_apply == Left) ? pCardValue : pSlotValue;
    Value* pArg = (move.m_apply == Left) ? pSlotValue : pCardValue;

    try
    {
        if (m_v[m_player][move.m_slot] <= 0)
        {
            throw Error();
        }

        m_applyCount = 0;
        Value* pResult = apply(file, pFunc, pArg);
        m_f[m_player][move.m_slot]->release();
        m_f[m_player][move.m_slot] = pResult;
    }
    catch (Error)
    {
        m_f[m_player][move.m_slot]->release();
        m_f[m_player][move.m_slot] = m_pI;
        m_pI->addRef();
    }

    pFunc->release();
    pArg->release();

    m_player = 1 - m_player;
    if (m_player == 0)
    {
        m_turn++;
    }
}

Value* State::apply(FILE* file, Value* pFunc, Value* pArg)
{
#if 0
    if (file)
    {
        fprintf(file, "Apply: ");
        pFunc->dump(file);
        fprintf(file, " -> ");
        pArg->dump(file);
        fprintf(file, "\n");
        fflush(file);
    }
#endif

    if (m_applyCount >= 1000)
    {
        throw Error();
    }
    m_applyCount++;

    switch (pFunc->m_type)
    {
    case VT_int:
    {
        throw Error();
    }
    break;
    case VT_I:
    {
        pArg->addRef();
        return pArg;
    }
    break;
    case VT_succ:
    {
        if (pArg->m_type != VT_int)
        {
            throw Error();
        }
        if (pArg->m_int == 65535)
        {
            pArg->addRef();
            return pArg;
        }
        Value* pResult = new Value();
        pResult->m_type = VT_int;
        pResult->m_int = pArg->m_int + 1;
        return pResult;
    }
    break;
    case VT_dbl:
    {
        if (pArg->m_type != VT_int)
        {
            throw Error();
        }
        if (pArg->m_int == 65535)
        {
            pArg->addRef();
            return pArg;
        }
        Value* pResult = new Value();
        pResult->m_type = VT_int;
        if (pArg->m_int < 32768)
        {
            pResult->m_int = pArg->m_int * 2;
        }
        else
        {
            pResult->m_int = 65535;
        }
        return pResult;
    }
    break;
    case VT_get:
    {
        if (pArg->m_type != VT_int ||
            pArg->m_int < 0 ||
            pArg->m_int > 255 ||
            m_v[m_player][pArg->m_int] <= 0)
        {
            throw Error();
        }

        Value* pResult = m_f[m_player][pArg->m_int];
        pResult->addRef();
        return pResult;
    }
    break;
    case VT_put:
    {
        m_pI->addRef();
        return m_pI;
    }
    break;
    case VT_S:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_S_1;

        pResult->m_arg1 = pArg;
        pResult->m_arg1->addRef();

        return pResult;
    }
    break;
    case VT_S_1:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_S_2;

        pResult->m_arg1 = pFunc->m_arg1;
        pResult->m_arg1->addRef();

        pResult->m_arg2 = pArg;
        pResult->m_arg2->addRef();

        return pResult;
    }
    break;
    case VT_S_2:
    {
        Value* pH = apply(file, pFunc->m_arg1, pArg);

        try
        {
            Value* pY = apply(file, pFunc->m_arg2, pArg);

            try
            {
                Value* pZ = apply(file, pH, pY);
                pH->release();
                pY->release();
                return pZ;
            }
            catch (...)
            {
                pY->release();
                throw;
            }
        }
        catch (...)
        {
            pH->release();
            throw;
        }
    }
    break;
    case VT_K:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_K_1;

        pResult->m_arg1 = pArg;
        pResult->m_arg1->addRef();

        return pResult;
    }
    break;
    case VT_K_1:
    {
        Value* pResult = pFunc->m_arg1;
        pResult->addRef();
        return pResult;
    }
    break;
    case VT_inc:
    {
        if (pArg->m_type != VT_int ||
            pArg->m_int < 0 ||
            pArg->m_int > 255)
        {
            throw Error();
        }

        if (m_zombie)
        {
            if (m_v[m_player][pArg->m_int] > 0)
            {
                m_v[m_player][pArg->m_int]--;
            }
        }
        else
        {
            if (m_v[m_player][pArg->m_int] > 0 &&
                m_v[m_player][pArg->m_int] < 65535)
            {
                m_v[m_player][pArg->m_int]++;
            }
        }

        m_pI->addRef();
        return m_pI;
    }
    break;
    case VT_dec:
    {
        if (pArg->m_type != VT_int ||
            pArg->m_int < 0 ||
            pArg->m_int > 255)
        {
            throw Error();
        }

        if (m_zombie)
        {
            if (m_v[1-m_player][255-pArg->m_int] > 0 &&
                m_v[1-m_player][255-pArg->m_int] < 65535)
            {
                m_v[1-m_player][255-pArg->m_int]++;
            }
        }
        else
        {
            if (m_v[1-m_player][255-pArg->m_int] > 0)
            {
                m_v[1-m_player][255-pArg->m_int]--;
            }
        }

        m_pI->addRef();
        return m_pI;
    }
    break;
    case VT_attack:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_attack_1;

        pResult->m_arg1 = pArg;
        pResult->m_arg1->addRef();

        return pResult;
    }
    break;
    case VT_attack_1:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_attack_2;

        pResult->m_arg1 = pFunc->m_arg1;
        pResult->m_arg1->addRef();

        pResult->m_arg2 = pArg;
        pResult->m_arg2->addRef();

        return pResult;
    }
    break;
    case VT_attack_2:
    {
        if (pFunc->m_arg1->m_type != VT_int ||
            pFunc->m_arg1->m_int < 0 ||
            pFunc->m_arg1->m_int > 255 ||
            pArg->m_type != VT_int ||
            pArg->m_int > m_v[m_player][pFunc->m_arg1->m_int])
        {
            throw Error();
        }

        m_v[m_player][pFunc->m_arg1->m_int] -= pArg->m_int;

        if (pFunc->m_arg2->m_type != VT_int ||
            pFunc->m_arg2->m_int < 0 ||
            pFunc->m_arg2->m_int > 255)
        {
            throw Error();
        }

        if (m_zombie)
        {
            if (m_v[1-m_player][255-pFunc->m_arg2->m_int] > 0)
            {
                m_v[1-m_player][255-pFunc->m_arg2->m_int] +=
                    ((pArg->m_int * 9) / 10);
                if (m_v[1-m_player][255-pFunc->m_arg2->m_int] > 65535)
                {
                    m_v[1-m_player][255-pFunc->m_arg2->m_int] = 65535;
                }
            }
        }
        else
        {
            if (m_v[1-m_player][255-pFunc->m_arg2->m_int] > 0)
            {
                m_v[1-m_player][255-pFunc->m_arg2->m_int] -=
                    ((pArg->m_int * 9) / 10);
                if (m_v[1-m_player][255-pFunc->m_arg2->m_int] < 0)
                {
                    m_v[1-m_player][255-pFunc->m_arg2->m_int] = 0;
                }
            }
        }

        m_pI->addRef();
        return m_pI;
    }
    break;
    case VT_help:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_help_1;

        pResult->m_arg1 = pArg;
        pResult->m_arg1->addRef();

        return pResult;
    }
    break;
    case VT_help_1:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_help_2;

        pResult->m_arg1 = pFunc->m_arg1;
        pResult->m_arg1->addRef();

        pResult->m_arg2 = pArg;
        pResult->m_arg2->addRef();

        return pResult;
    }
    break;
    case VT_help_2:
    {
        if (pFunc->m_arg1->m_type != VT_int ||
            pFunc->m_arg1->m_int < 0 ||
            pFunc->m_arg1->m_int > 255 ||
            pArg->m_type != VT_int ||
            pArg->m_int > m_v[m_player][pFunc->m_arg1->m_int])
        {
            throw Error();
        }

        m_v[m_player][pFunc->m_arg1->m_int] -= pArg->m_int;

        if (pFunc->m_arg2->m_type != VT_int ||
            pFunc->m_arg2->m_int < 0 ||
            pFunc->m_arg2->m_int > 255)
        {
            throw Error();
        }

        if (m_zombie)
        {
            if (m_v[m_player][pFunc->m_arg2->m_int] > 0)
            {
                m_v[m_player][pFunc->m_arg2->m_int] -=
                    ((pArg->m_int * 11) / 10);
                if (m_v[m_player][pFunc->m_arg2->m_int] < 0)
                {
                    m_v[m_player][pFunc->m_arg2->m_int] = 0;
                }
            }
        }
        else
        {
            if (m_v[m_player][pFunc->m_arg2->m_int] > 0)
            {
                m_v[m_player][pFunc->m_arg2->m_int] +=
                    ((pArg->m_int * 11) / 10);
                if (m_v[m_player][pFunc->m_arg2->m_int] > 65535)
                {
                    m_v[m_player][pFunc->m_arg2->m_int] = 65535;
                }
            }
        }

        m_pI->addRef();
        return m_pI;
    }
    break;
    case VT_copy:
    {
        if (pArg->m_type != VT_int ||
            pArg->m_int < 0 ||
            pArg->m_int > 255)
        {
            throw Error();
        }

        Value *pResult = m_f[1-m_player][pArg->m_int];
        pResult->addRef();
        return pResult;
    }
    break;
    case VT_revive:
    {
        if (pArg->m_type != VT_int ||
            pArg->m_int < 0 ||
            pArg->m_int > 255)
        {
            throw Error();
        }

        if (m_v[m_player][pArg->m_int] <= 0)
        {
            m_v[m_player][pArg->m_int] = 1;
        }

        m_pI->addRef();
        return m_pI;
    }
    break;
    case VT_zombie:
    {
        Value* pResult = new Value();
        pResult->m_type = VT_zombie_1;

        pResult->m_arg1 = pArg;
        pResult->m_arg1->addRef();

        return pResult;
    }
    break;
    case VT_zombie_1:
    {
        if (pFunc->m_arg1->m_type != VT_int ||
            pFunc->m_arg1->m_int < 0 ||
            pFunc->m_arg1->m_int > 255 ||
            m_v[1-m_player][255-pFunc->m_arg1->m_int] > 0)
        {
            throw Error();
        }

        m_f[1-m_player][255-pFunc->m_arg1->m_int]->release();
        pArg->addRef();
        m_f[1-m_player][255-pFunc->m_arg1->m_int] = pArg;

        m_v[1-m_player][255-pFunc->m_arg1->m_int] = -1;

        m_pI->addRef();
        return m_pI;
    }
    break;
    }

    throw Error();
}

void State::dump(FILE* file) const
{
    fprintf(file, "Player %d's turn, with slots:\n", m_player);

    for (int slot = 0; slot < 256; slot++)
    {
        if (m_v[m_player][slot] != 10000 ||
            m_f[m_player][slot]->m_type != VT_I)
        {
            fprintf(file, "%d={%d,", slot, m_v[m_player][slot]);
            m_f[m_player][slot]->dump(file);
            fprintf(file, "}\n");
        }
    }

    fprintf(file, "(slots {10000,I} are omitted)\n");
}
