#ifndef STATE_H
#define STATE_H

#include "Move.h"
#include <stdio.h>
#include <vector>

enum ValueType
{
    VT_int,
    VT_I,
    VT_succ,
    VT_dbl,
    VT_get,
    VT_put,
    VT_S,
    VT_S_1,
    VT_S_2,
    VT_K,
    VT_K_1,
    VT_inc,
    VT_dec,
    VT_attack,
    VT_attack_1,
    VT_attack_2,
    VT_help,
    VT_help_1,
    VT_help_2,
    VT_copy,
    VT_revive,
    VT_zombie,
    VT_zombie_1
};

class Value
{
public:
    Value();
    Value(Card card);
    ~Value();

    void addRef() { m_refCount++; }
    void release() { m_refCount--; if (m_refCount == 0) delete this; }

    bool equals(const Value* pOther) const;

    void dump(FILE* file) const;

    ValueType m_type;
    Value* m_arg1;
    Value* m_arg2;
    int m_int;

    int m_refCount;
    static int s_allocCount;
};

class ValuePtr
{
public:
    ValuePtr() : m_p(0) {}
    ~ValuePtr() { if (m_p) m_p->release(); }

    operator Value*() const { return m_p; }
    Value* operator->() const { return m_p; }

    Value* m_p;
};

class State
{
public:
    State();
    ~State();

    State(const State& other);
    State& operator=(const State& other);

    void preStep(FILE* file, std::vector<int>* pZombies);
    void step(FILE* file, Move move);

    Value* apply(FILE* file, Value* pFunc, Value* pArg);

    void dump(FILE* file) const;

    int m_player;
    int m_turn;
    int m_v[2][256];
    Value* m_f[2][256];
    int m_applyCount;
    bool m_zombie;
    Value* m_pI;
};

#endif
