#include "Move.h"
#include "Program.h"
#include "Build.h"
#include <vector>
#include <stdlib.h>

using std::vector;

const int zombieSlot = 2;
const int zombieAmountSlot = 3;

struct ProgState
{
    enum Enum
    {
        Start,
        Build,
        Revive,
        AfterBuildRevive,
        AfterZombie,
        ZombieAttack,
        AfterZombieAmount,
        AfterZombieProg,
        AfterZombieGo,
        Decide,
        Nop
    };
};

class Prog : public Program
{
public:
    Prog();
    ~Prog();

    virtual void opponentZombies(std::vector<int>& zombies);
    virtual void proponent(Move* pMove);

    Value* m_pFenceValue;
    Value* m_pZombieValue;

    ProgState::Enum m_ps;
    vector<Move> m_moves;
    int m_idx;
    vector<int> m_buildNeeds;
    bool m_buildFailed;
    ProgState::Enum m_nextState;
    int m_buildReviveFailCount;

    int m_reviveDeadSlot;
    int m_reviveLiveSlot;
    int m_reviveNum;
    ProgState::Enum m_afterRevive;
    bool m_reviveFailed;

    bool m_zombieFailed;
    bool m_zombieBuilt;

    int m_zombieAttackFailCount;
    int m_zombieStart;
    int m_zombieAmount;
    bool m_zombieAmountFailed;
    bool m_zombieAmountBuilt;
    bool m_zombieProgFailed;
    bool m_zombieProgBuilt;
};

Prog::Prog()
{
    m_pFenceValue =
        _S(
            _S(
                _S(
                    _S(help, succ),
                    _K(9216)
                    ),
                _S(_K(get), _K(1))
                ),
            _S(_K(succ), succ)
            )
        ;

    m_pZombieValue =
        _S(
            _S(
                _S(
                    _S(help, I),
                    _S(_K(copy), _K(zombieAmountSlot))
                    ),
                _S(_K(copy), _K(zombieSlot))
                ),
            succ
            )
        ;

    m_zombieFailed = false;
    m_zombieAttackFailCount = 0;

    m_ps = ProgState::Start;
}

Prog::~Prog()
{
    m_pFenceValue->release();
}

void Prog::opponentZombies(std::vector<int>& zombies)
{
    if (m_ps == ProgState::Build)
    {
        for (int i = 0; i < zombies.size(); i++)
        {
            for (int j = 0; j < m_buildNeeds.size(); j++)
            {
                if (zombies[i] == m_buildNeeds[j])
                {
                    m_ps = m_nextState;
                    m_buildFailed = true;
                }
            }
        }
    }
}

void Prog::proponent(Move* pMove)
{
    while (true)
    {
        switch (m_ps)
        {

        case ProgState::Start:
        {
            int buildSlot = 1;

            BuildResult buildResult;
            getBuild(m_pFenceValue, &buildResult, buildSlot, 0);

            m_moves = buildResult.m_moves[true][true];
            m_idx = 0;

            m_buildNeeds.clear();
            m_buildNeeds.push_back(buildSlot);
            if (buildResult.m_usesZero[true])
            {
                m_buildNeeds.push_back(0);
            }

            m_moves.insert(
                m_moves.begin(),
                Move(put, buildSlot));

            m_moves.push_back(
                Move(buildSlot, zero));

            m_nextState = ProgState::Decide;
            m_buildFailed = false;
            m_buildReviveFailCount = 0;

            m_ps = ProgState::Build;
        }
        break;

        case ProgState::Revive:
        {
            // Pick a live slot
            if (m_reviveLiveSlot == -1)
            {
                int live[256];
                int liveCount = 0;
                for (int pass = 1; pass <= 3; pass++)
                {
                    liveCount = 0;
                    for (int slot = 0; slot < 256; slot++)
                    {
                        if (pass == 1)
                        {
                            bool skip = false;
                            if (m_afterRevive == ProgState::AfterBuildRevive)
                            {
                                for (int j = 0; j < m_buildNeeds.size(); j++)
                                {
                                    if (m_buildNeeds[j] == slot)
                                    {
                                        skip = true;
                                        break;
                                    }
                                }
                            }
                            if (skip)
                            {
                                continue;
                            }
                        }

                        if (((pass < 3 && slot == zombieSlot) ||
                             (pass == 1 && slot == zombieAmountSlot)) &&
                            !m_zombieFailed &&
                            (!m_zombieBuilt ||
                             m_state.m_f[m_player][zombieSlot]->equals(m_pZombieValue)))
                        {
                            continue;
                        }
                        if (m_state.m_v[m_player][slot] > 0)
                        {
                            live[liveCount++] = slot;
                        }
                    }

                    if (liveCount > 0)
                    {
                        break;
                    }
                }

                if (liveCount == 0)
                {
                    m_ps = m_afterRevive;
                    m_reviveFailed = true;
                    continue;
                }

                m_reviveLiveSlot = live[rand() % liveCount];
                if (m_state.m_f[m_player][m_reviveLiveSlot]->m_type == VT_I)
                {
                    m_reviveNum = -1;
                }
                else
                {
                    m_reviveNum = -2;
                }
            }

            // Check if anything has impeded our progress
            bool problem = false;
            if (m_state.m_v[m_player][m_reviveLiveSlot] <= 0 ||
                m_state.m_v[m_player][m_reviveDeadSlot] > 0)
            {
                problem = true;
            }
            else
            {
                if (m_reviveNum == -1)
                {
                    if (m_state.m_f[m_player][m_reviveLiveSlot]->m_type !=
                        VT_I)
                    {
                        problem = true;
                    }
                }
                else if (m_reviveNum >= 0)
                {
                    if (m_state.m_f[m_player][m_reviveLiveSlot]->m_type !=
                        VT_int ||
                        m_state.m_f[m_player][m_reviveLiveSlot]->m_int !=
                        m_reviveNum)
                    {
                        problem = true;
                    }
                }
            }
            if (problem)
            {
                m_ps = m_afterRevive;
                m_reviveFailed = true;
                continue;
            }

            // Do the next step
            if (m_reviveNum == -2)
            {
                *pMove = Move(put, m_reviveLiveSlot);
                m_reviveNum = -1;
                return;
            }
            else if (m_reviveNum == -1)
            {
                *pMove = Move(m_reviveLiveSlot, zero);
                m_reviveNum = 0;
                return;
            }
            else if (m_reviveNum == m_reviveDeadSlot)
            {
                *pMove = Move(revive, m_reviveLiveSlot);
                m_ps = m_afterRevive;
                return;
            }
            else if (m_reviveNum == 0)
            {
                *pMove = Move(succ, m_reviveLiveSlot);
                m_reviveNum = 1;
                return;
            }
            else
            {
                int goalPos = 15;
                while ((m_reviveDeadSlot & (1 << goalPos)) == 0)
                {
                    goalPos--;
                }

                int curPos = 15;
                while ((m_reviveNum & (1 << curPos)) == 0)
                {
                    curPos--;
                }

                while (curPos >= 0 &&
                       ((m_reviveDeadSlot & (1 << goalPos)) != 0) ==
                       ((m_reviveNum & (1 << curPos)) != 0))
                {
                    goalPos--;
                    curPos--;
                }

                if (curPos == -1)
                {
                    *pMove = Move(dbl, m_reviveLiveSlot);
                    m_reviveNum *= 2;
                }
                else
                {
                    *pMove = Move(succ, m_reviveLiveSlot);
                    m_reviveNum++;
                }
                return;
            }
        }
        break;

        case ProgState::AfterBuildRevive:
        {
            if (m_reviveFailed)
            {
                m_buildReviveFailCount++;
            }
            m_ps = ProgState::Build;
        }
        break;

        case ProgState::Build:
        {
            if (m_buildReviveFailCount >= 3)
            {
                m_ps = m_nextState;
                m_buildFailed = true;
                continue;
            }

            for (int i = 0; i < m_buildNeeds.size(); i++)
            {
                if (m_state.m_v[m_player][m_buildNeeds[i]] <= 0)
                {
                    m_ps = ProgState::Revive;
                    m_reviveDeadSlot = m_buildNeeds[i];
                    m_reviveLiveSlot = -1;
                    m_afterRevive = ProgState::AfterBuildRevive;
                    m_reviveFailed = false;
                    break;
                }
            }

            if (m_ps != ProgState::Build)
            {
                continue;
            }

            if (m_nextState == ProgState::AfterZombieAmount ||
                m_nextState == ProgState::AfterZombieProg ||
                m_nextState == ProgState::AfterZombieGo)
            {
                if (m_state.m_v[1-m_player][255-1] > 1)
                {
                    m_ps = m_nextState;
                    m_buildFailed = true;
                    continue;
                }
            }

            while (m_ps == ProgState::Build &&
                   m_moves[m_idx].m_apply == Left &&
                   m_moves[m_idx].m_card == put &&
                   m_state.m_f[m_player][m_moves[m_idx].m_slot]->m_type ==
                   VT_I)
            {
                m_idx++;
                if (m_idx == m_moves.size())
                {
                    m_ps = m_nextState;
                }
            }

            if (m_ps != ProgState::Build)
            {
                continue;
            }

            *pMove = m_moves[m_idx];
            m_idx++;
            m_buildReviveFailCount = 0;
            if (m_idx == m_moves.size())
            {
                m_ps = m_nextState;
            }
            return;
        }
        break;

        case ProgState::AfterZombie:
        {
            if (m_buildFailed)
            {
                m_zombieFailed = true;
            }
            else
            {
                m_zombieBuilt = true;
            }
            m_ps = ProgState::Decide;
        }
        break;

        case ProgState::Decide:
        {
            // Don't bother detected too many failures since I never
            // implemented any fallback strategies
            {
                if (m_zombieFailed)
                {
                    m_zombieFailed = false;
                }
                if (m_zombieBuilt &&
                    !m_state.m_f[m_player][zombieSlot]->equals(m_pZombieValue))
                {
                    m_zombieBuilt = false;
                }
                m_zombieAttackFailCount = 0;
            }

            if (!m_zombieFailed &&
                !m_zombieBuilt)
            {
                int buildSlot = zombieSlot;

                BuildResult buildResult;
                getBuild(m_pZombieValue, &buildResult, buildSlot, 0);

                m_moves = buildResult.m_moves[true][true];
                m_idx = 0;

                m_buildNeeds.clear();
                m_buildNeeds.push_back(buildSlot);
                if (buildResult.m_usesZero[true])
                {
                    m_buildNeeds.push_back(0);
                }

                m_moves.insert(
                    m_moves.begin(),
                    Move(put, buildSlot));

                m_nextState = ProgState::AfterZombie;
                m_buildFailed = false;
                m_buildReviveFailCount = 0;

                m_ps = ProgState::Build;
                continue;
            }

            if (m_zombieBuilt &&
                m_state.m_f[m_player][zombieSlot]->equals(m_pZombieValue) &&
                m_zombieAttackFailCount < 3 &&
                m_state.m_v[1-m_player][255-1] > 1 /*0*/ /*100*/)
            {
                int maxAttack[256];
                for (int slot = 0; slot < 256; slot++)
                {
                    maxAttack[slot] = 0;
                    for (int i = 0; i < 15; i++)
                    {
                        if (m_state.m_v[m_player][slot] > (1<<i))
                        {
                            maxAttack[slot] = (1<<i);
                        }
                        for (int j = 0; j < i; j++)
                        {
                            if (m_state.m_v[m_player][slot] > ((1<<i)|(1<<j)))
                            {
                                maxAttack[slot] = ((1<<i)|(1<<j));
                            }
                        }
                    }
                }

                int neededAttack = (m_state.m_v[1-m_player][255-1]*10+8)/9;
                int maxMaxAttack = 0;
                for (int slot = 0; slot < 256; slot++)
                {
                    if (maxAttack[slot] > maxMaxAttack)
                    {
                        maxMaxAttack = maxAttack[slot];
                    }
                }

                int oneList[256];
                int oneCount = 0;
                int maxList[256];
                int maxCount = 0;
                for (int slot = 0; slot < 256; slot++)
                {
                    if (maxAttack[slot] >= neededAttack)
                    {
                        oneList[oneCount++] = slot;
                    }
                    if (maxAttack[slot] == maxMaxAttack)
                    {
                        maxList[maxCount++] = slot;
                    }
                }

                int chosenSlot;
                int chosenAttack = 0;
                if (oneCount > 0)
                {
                    chosenSlot = oneList[rand() % oneCount];
                    for (int i = 0; i < 15; i++)
                    {
                        chosenAttack = (1<<i);
                        if (chosenAttack >= neededAttack)
                        {
                            break;
                        }
                        for (int j = 0; j < i; j++)
                        {
                            chosenAttack = ((1<<i)|(1<<j));
                            if (chosenAttack >= neededAttack)
                            {
                                break;
                            }
                        }
                        if (chosenAttack >= neededAttack)
                        {
                            break;
                        }
                    }
                }
                else
                {
                    chosenSlot = maxList[rand() % maxCount];
                    chosenAttack = maxMaxAttack;
                }

                int buildSlot = 32 + (rand() % 64);
                Value* pValue1 = _S(_K(_attack(chosenSlot, 1)), get);
                Value* pValue2 = _(chosenAttack);

                BuildResult buildResult1;
                getBuild(pValue1, &buildResult1, buildSlot, 0);

                BuildResult buildResult2;
                getBuild(pValue2, &buildResult2, buildSlot, 0);

                pValue1->release();
                pValue2->release();

                m_moves = buildResult1.m_moves[true][true];
                m_moves.insert(
                    m_moves.begin(),
                    Move(put, buildSlot));
                m_moves.push_back(
                    Move(put, 0));
                m_moves.insert(
                    m_moves.end(),
                    buildResult2.m_moves[true][false].begin(),
                    buildResult2.m_moves[true][false].end());
                m_moves.push_back(
                    Move(buildSlot, zero));
                m_idx = 0;

                m_buildNeeds.clear();
                m_buildNeeds.push_back(buildSlot);
                m_buildNeeds.push_back(0);

                m_nextState = ProgState::Decide;
                m_buildFailed = false;
                m_buildReviveFailCount = 0;

                m_ps = ProgState::Build;
                continue;
            }

            if (m_zombieBuilt &&
                m_state.m_f[m_player][zombieSlot]->equals(m_pZombieValue) &&
                m_zombieAttackFailCount < 3 &&
                m_state.m_v[1-m_player][255-1] <= 1)
            {
                int twoList[256];
                int twoCount = 0;
                int oneList[256];
                int oneCount = 0;
                for (int slot = 0; slot < 256; slot++)
                {
                    int v = m_state.m_v[1-m_player][slot];
                    if (v >= 2)
                    {
                        twoList[twoCount++] = slot;
                    }
                    else if (v >= 1)
                    {
                        oneList[oneCount++] = slot;
                    }
                }

                if (twoCount > 0 ||
                    oneCount > 0)
                {
                    int seedSlot;
                    if (twoCount > 0)
                    {
                        seedSlot = twoList[rand() % twoCount];
                    }
                    else
                    {
                        seedSlot = oneList[rand() % oneCount];
                    }

                    int maxAmount = m_state.m_v[1-m_player][seedSlot];
                    int minAmount = (maxAmount * 10 + 20) / 21;

                    int fromSlot = seedSlot;
                    int toSlot = seedSlot;
                    int firstDir = rand() % 2;
                    for (int iDir = 0; iDir < 2; iDir++)
                    {
                        int dir = (iDir == 0) ? firstDir : 1-firstDir;

                        while (true)
                        {
                            if (toSlot - fromSlot + 1 >= 55)
                            {
                                break;
                            }

                            int curSlot =
                                (dir == 0) ? toSlot + 1 : fromSlot - 1;
                            if (curSlot < 0 || curSlot > 255)
                            {
                                break;
                            }

                            int curMax = m_state.m_v[1-m_player][curSlot];
                            int curMin = (curMax * 10 + 20) / 21;
                            if (curMax < minAmount ||
                                curMin > maxAmount)
                            {
                                break;
                            }

                            if (curMax < maxAmount)
                            {
                                maxAmount = curMax;
                            }
                            if (curMin > minAmount)
                            {
                                minAmount = curMin;
                            }

                            if (dir == 0)
                            {
                                toSlot = curSlot;
                            }
                            else
                            {
                                fromSlot = curSlot;
                            }
                        }
                    }

                    m_zombieStart = fromSlot;
                    m_zombieAmount = minAmount;
                    m_zombieAmountFailed = false;
                    m_zombieAmountBuilt = false;
                    m_zombieProgFailed = false;
                    m_zombieProgBuilt = false;
                    m_ps = ProgState::ZombieAttack;
                    continue;
                }
            }

            m_ps = ProgState::Nop;
            continue;

            //int zombieSlot = 255;
            //while (zombieSlot >= 0 &&
        }
        break;

        case ProgState::ZombieAttack:
        {
            if (!m_state.m_f[m_player][zombieSlot]->equals(m_pZombieValue))
            {
                m_ps = ProgState::Decide;
            }

            if (!m_zombieAmountFailed &&
                !m_zombieAmountBuilt)
            {
                int buildSlot = zombieAmountSlot;

                Value* pValue = _(m_zombieAmount);

                BuildResult buildResult;
                getBuild(pValue, &buildResult, buildSlot, 0);

                pValue->release();

                m_moves = buildResult.m_moves[true][true];
                m_idx = 0;

                m_buildNeeds.clear();
                m_buildNeeds.push_back(buildSlot);
                if (buildResult.m_usesZero[true])
                {
                    m_buildNeeds.push_back(0);
                }

                m_moves.insert(
                    m_moves.begin(),
                    Move(put, buildSlot));

                m_nextState = ProgState::AfterZombieAmount;
                m_buildFailed = false;
                m_buildReviveFailCount = 0;

                m_ps = ProgState::Build;
                continue;
            }

            if (m_zombieAmountFailed)
            {
                m_zombieAttackFailCount++;
                m_ps = ProgState::Decide;
                continue;
            }

            if (m_state.m_f[m_player][zombieAmountSlot]->m_type !=
                VT_int ||
                m_state.m_f[m_player][zombieAmountSlot]->m_int !=
                m_zombieAmount)
            {
                m_zombieAttackFailCount++;
                m_ps = ProgState::Decide;
                continue;
            }

            if (!m_zombieProgFailed &&
                !m_zombieProgBuilt)
            {
                int buildSlot = 1;

                Value* pValue = _K(m_zombieStart);

                BuildResult buildResult;
                getBuild(pValue, &buildResult, buildSlot, 0);

                pValue->release();

                m_moves.clear();
                m_moves.push_back(
                    Move(put, buildSlot));
                m_moves.push_back(
                    Move(buildSlot, zero));
                m_moves.push_back(
                    Move(succ, buildSlot));
                m_moves.push_back(
                    Move(dbl, buildSlot));
                m_moves.push_back(
                    Move(get, buildSlot));
                m_moves.push_back(
                    Move(K, buildSlot));
                m_moves.push_back(
                    Move(S, buildSlot));
                m_moves.push_back(
                    Move(K, buildSlot));
                m_moves.push_back(
                    Move(S, buildSlot));
                m_moves.push_back(
                    Move(buildSlot, get));
                m_moves.push_back(
                    Move(put, 0));
                m_moves.insert(
                    m_moves.end(),
                    buildResult.m_moves[true][false].begin(),
                    buildResult.m_moves[true][false].end());
                m_moves.push_back(
                    Move(buildSlot, zero));

                m_idx = 0;

                m_buildNeeds.clear();
                m_buildNeeds.push_back(buildSlot);
                m_buildNeeds.push_back(0);
                m_buildNeeds.push_back(2);

                m_nextState = ProgState::AfterZombieProg;
                m_buildFailed = false;
                m_buildReviveFailCount = 0;

                m_ps = ProgState::Build;
                continue;
            }

            if (m_zombieProgFailed)
            {
                m_zombieAttackFailCount++;
                m_ps = ProgState::Decide;
                continue;
            }

            if (true)
            {
                int buildSlot = 32 + (rand() % 64);

                m_moves.clear();
                m_moves.push_back(
                    Move(put, buildSlot));
                m_moves.push_back(
                    Move(buildSlot, dec));
                m_moves.push_back(
                    Move(S, buildSlot));
                m_moves.push_back(
                    Move(buildSlot, zombie));
                m_moves.push_back(
                    Move(S, buildSlot));
                m_moves.push_back(
                    Move(buildSlot, get));
                m_moves.push_back(
                    Move(K, buildSlot));
                m_moves.push_back(
                    Move(S, buildSlot));
                m_moves.push_back(
                    Move(buildSlot, succ));
                m_moves.push_back(
                    Move(buildSlot, zero));

                m_idx = 0;

                m_buildNeeds.clear();
                m_buildNeeds.push_back(buildSlot);
                m_buildNeeds.push_back(1);

                m_nextState = ProgState::AfterZombieGo;
                m_buildFailed = false;
                m_buildReviveFailCount = 0;

                m_ps = ProgState::Build;
                continue;
            }

            m_ps = ProgState::Decide;
        }
        break;

        case ProgState::AfterZombieAmount:
        {
            if (m_buildFailed)
            {
                m_zombieAmountFailed = true;
            }
            else
            {
                m_zombieAmountBuilt = true;
            }
            m_ps = ProgState::ZombieAttack;
        }
        break;

        case ProgState::AfterZombieProg:
        {
            if (m_buildFailed)
            {
                m_zombieProgFailed = true;
            }
            else
            {
                m_zombieProgBuilt = true;
            }
            m_ps = ProgState::ZombieAttack;
        }
        break;

        case ProgState::AfterZombieGo:
        {
            if (m_buildFailed)
            {
                m_zombieAttackFailCount++;
            }
            else
            {
                m_zombieAttackFailCount = 0;
            }
            m_ps = ProgState::Decide;
        }
        break;

        case ProgState::Nop:
        default:
        {
            int slot = 0;
            while (slot < 255 &&
                   m_state.m_v[m_player][slot] <= 0)
            {
                slot++;
            }

            *pMove = Move(I, slot);
            m_ps = ProgState::Decide;
            return;
        }
        }
    }
}

int main(int argc, char* argv[])
{
    Prog program;
    return program.run(argc, argv);
}
