#include "Move.h"
#include "Program.h"
#include <stdio.h>
#include <stdlib.h>

class Prog : public Program
{
public:
    Prog() : m_liveSlot(-1), m_deadSlot(-1), m_num(-1) {}

    virtual void proponent(Move* pMove);

    int m_liveSlot;
    int m_deadSlot;
    int m_num;
};

void Prog::proponent(Move* pMove)
{
    if (m_deadSlot != -1)
    {
        // If something has impeded our progress, start over
        if (m_state.m_v[m_player][m_liveSlot] <= 0 ||
            m_state.m_v[m_player][m_deadSlot] > 0)
        {
            m_deadSlot = -1;
        }
        else
        {
            if (m_num == -1)
            {
                if (m_state.m_f[m_player][m_liveSlot]->m_type != VT_I)
                {
                    m_deadSlot = -1;
                }
            }
            else
            {
                if (m_state.m_f[m_player][m_liveSlot]->m_type != VT_int ||
                    m_state.m_f[m_player][m_liveSlot]->m_int != m_num)
                {
                    m_deadSlot = -1;
                }
            }
        }
    }

    if (m_deadSlot == -1)
    {
        // Try to pick a new dead slot
        int live[256];
        int dead[256];
        int liveCount = 0;
        int deadCount = 0;
        for (int slot = 0; slot < 256; slot++)
        {
            if (m_state.m_v[m_player][slot] > 0)
            {
                live[liveCount++] = slot;
            }
            else
            {
                dead[deadCount++] = slot;
            }
        }

        if (deadCount > 0 && liveCount > 0)
        {
            m_deadSlot = dead[rand() % deadCount];
            m_liveSlot = live[rand() % liveCount];
            if (m_state.m_f[m_player][m_liveSlot]->m_type = VT_I)
            {
                m_num = -1;
            }
            else
            {
                m_num = -2;
            }
        }
    }

    if (m_deadSlot != -1)
    {
        // Do the next step
        if (m_num == -2)
        {
            *pMove = Move(put, m_liveSlot);
            m_num = -1;
            return;
        }
        else if (m_num == -1)
        {
            *pMove = Move(m_liveSlot, zero);
            m_num = 0;
            return;
        }
        else if (m_num == m_deadSlot)
        {
            *pMove = Move(revive, m_liveSlot);
            m_deadSlot = -1;
            return;
        }
        else if (m_num == 0)
        {
            *pMove = Move(succ, m_liveSlot);
            m_num = 1;
            return;
        }
        else
        {
            int goalPos = 15;
            while ((m_deadSlot & (1 << goalPos)) == 0)
            {
                goalPos--;
            }

            int curPos = 15;
            while ((m_num & (1 << curPos)) == 0)
            {
                curPos--;
            }

            while (curPos >= 0 &&
                   ((m_deadSlot & (1 << goalPos)) != 0) ==
                   ((m_num & (1 << curPos)) != 0))
            {
                goalPos--;
                curPos--;
            }

            if (curPos == -1)
            {
                *pMove = Move(dbl, m_liveSlot);
                m_num *= 2;
            }
            else
            {
                *pMove = Move(succ, m_liveSlot);
                m_num++;
            }
            return;
        }
    }

    *pMove = Move(I, 0);
}

int main(int argc, char* argv[])
{
    Prog program;
    return program.run(argc, argv);
}
